package edu.princeton.swing;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.undo.*;

/**
 * ActionedUndoManager extends UndoManager to provide usable Action objects for the undo and
 * redo properties.
 */

public class ActionedUndoManager extends UndoManager
{
    private static final Toolkit TOOLKIT = Toolkit.getDefaultToolkit();

    /**
     * The bytes for the undo icon.
     */
    public static final byte UNDO_ICON_BYTES[] =
    {
        (byte)0x47, (byte)0x49, (byte)0x46, (byte)0x38, (byte)0x39, (byte)0x61,
        (byte)0x18, (byte)0x00, (byte)0x18, (byte)0x00, (byte)0xA2, (byte)0xFF,
        (byte)0x00, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xCC, (byte)0xCC,
        (byte)0xFF, (byte)0x99, (byte)0x99, (byte)0xCC, (byte)0x33, (byte)0x33,
        (byte)0x66, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xC0, (byte)0xC0,
        (byte)0xC0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x21, (byte)0xF9, (byte)0x04, (byte)0x01, (byte)0x00,
        (byte)0x00, (byte)0x05, (byte)0x00, (byte)0x2C, (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x00, (byte)0x18, (byte)0x00, (byte)0x18, (byte)0x00,
        (byte)0x00, (byte)0x03, (byte)0x61, (byte)0x58, (byte)0xBA, (byte)0xDC,
        (byte)0xFE, (byte)0x30, (byte)0xCA, (byte)0x42, (byte)0xAA, (byte)0x9D,
        (byte)0xAD, (byte)0x86, (byte)0xCD, (byte)0x2B, (byte)0xA6, (byte)0x9B,
        (byte)0x20, (byte)0x8E, (byte)0x1B, (byte)0x31, (byte)0x11, (byte)0xC1,
        (byte)0xA8, (byte)0x92, (byte)0x04, (byte)0x20, (byte)0x11, (byte)0xA3,
        (byte)0x05, (byte)0xAB, (byte)0xED, (byte)0x2B, (byte)0xB4, (byte)0x40,
        (byte)0x3E, (byte)0x8B, (byte)0xF5, (byte)0xEB, (byte)0x32, (byte)0xBB,
        (byte)0xDB, (byte)0xEF, (byte)0xB3, (byte)0xD8, (byte)0xF5, (byte)0x88,
        (byte)0x0A, (byte)0x14, (byte)0x6F, (byte)0x88, (byte)0x54, (byte)0x0A,
        (byte)0x31, (byte)0x83, (byte)0x4C, (byte)0xEA, (byte)0xF9, (byte)0x1A,
        (byte)0x0C, (byte)0x9D, (byte)0x54, (byte)0x08, (byte)0xCA, (byte)0x9A,
        (byte)0x9C, (byte)0x66, (byte)0x1F, (byte)0xCA, (byte)0x81, (byte)0xC5,
        (byte)0xFB, (byte)0x95, (byte)0x8A, (byte)0x38, (byte)0xE4, (byte)0xA3,
        (byte)0x03, (byte)0xBB, (byte)0x5A, (byte)0x9E, (byte)0xDA, (byte)0x3C,
        (byte)0x75, (byte)0x64, (byte)0x26, (byte)0xB6, (byte)0xE4, (byte)0x9A,
        (byte)0x02, (byte)0x2E, (byte)0x32, (byte)0xA3, (byte)0xDF, (byte)0x33,
        (byte)0xFA, (byte)0x7E, (byte)0x7E, (byte)0x09, (byte)0x00, (byte)0x21,
        (byte)0xFE, (byte)0x4F, (byte)0x43, (byte)0x6F, (byte)0x70, (byte)0x79,
        (byte)0x72, (byte)0x69, (byte)0x67, (byte)0x68, (byte)0x74, (byte)0x20,
        (byte)0x32, (byte)0x30, (byte)0x30, (byte)0x30, (byte)0x20, (byte)0x62,
        (byte)0x79, (byte)0x20, (byte)0x53, (byte)0x75, (byte)0x6E, (byte)0x20,
        (byte)0x4D, (byte)0x69, (byte)0x63, (byte)0x72, (byte)0x6F, (byte)0x73,
        (byte)0x79, (byte)0x73, (byte)0x74, (byte)0x65, (byte)0x6D, (byte)0x73,
        (byte)0x2C, (byte)0x20, (byte)0x49, (byte)0x6E, (byte)0x63, (byte)0x2E,
        (byte)0x20, (byte)0x41, (byte)0x6C, (byte)0x6C, (byte)0x20, (byte)0x52,
        (byte)0x69, (byte)0x67, (byte)0x68, (byte)0x74, (byte)0x73, (byte)0x20,
        (byte)0x52, (byte)0x65, (byte)0x73, (byte)0x65, (byte)0x72, (byte)0x76,
        (byte)0x65, (byte)0x64, (byte)0x2E, (byte)0x0D, (byte)0x0A, (byte)0x4A,
        (byte)0x4C, (byte)0x46, (byte)0x20, (byte)0x47, (byte)0x52, (byte)0x20,
        (byte)0x56, (byte)0x65, (byte)0x72, (byte)0x20, (byte)0x31, (byte)0x2E,
        (byte)0x30, (byte)0x0D, (byte)0x0A, (byte)0x00, (byte)0x3B
    };

    /**
     * The undo icon.
     */
    public static final Icon UNDO_ICON = new ImageIcon(
        TOOLKIT.createImage(UNDO_ICON_BYTES),
        "Undo"
        );

    /**
     * The bytes for the redo icon.
     */
    public static final byte REDO_ICON_BYTES[] =
    {
        (byte)0x47, (byte)0x49, (byte)0x46, (byte)0x38, (byte)0x39, (byte)0x61,
        (byte)0x18, (byte)0x00, (byte)0x18, (byte)0x00, (byte)0xA2, (byte)0xFF,
        (byte)0x00, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xCC, (byte)0xCC,
        (byte)0xFF, (byte)0x99, (byte)0x99, (byte)0xCC, (byte)0x33, (byte)0x33,
        (byte)0x66, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xC0, (byte)0xC0,
        (byte)0xC0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x21, (byte)0xF9, (byte)0x04, (byte)0x01, (byte)0x00,
        (byte)0x00, (byte)0x05, (byte)0x00, (byte)0x2C, (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x00, (byte)0x18, (byte)0x00, (byte)0x18, (byte)0x00,
        (byte)0x00, (byte)0x03, (byte)0x65, (byte)0x58, (byte)0xBA, (byte)0xDC,
        (byte)0xFE, (byte)0xB0, (byte)0x8C, (byte)0x48, (byte)0xD9, (byte)0x20,
        (byte)0xB5, (byte)0x8E, (byte)0x80, (byte)0x33, (byte)0x23, (byte)0x20,
        (byte)0xB1, (byte)0x09, (byte)0x5D, (byte)0x06, (byte)0x06, (byte)0x68,
        (byte)0x2A, (byte)0x90, (byte)0x26, (byte)0xBA, (byte)0xBE, (byte)0x70,
        (byte)0xF9, (byte)0x10, (byte)0x01, (byte)0x6C, (byte)0x93, (byte)0x40,
        (byte)0x44, (byte)0xC7, (byte)0xE0, (byte)0xB0, (byte)0x12, (byte)0x39,
        (byte)0x5D, (byte)0xED, (byte)0x07, (byte)0x28, (byte)0xFA, (byte)0x80,
        (byte)0x94, (byte)0x1D, (byte)0x71, (byte)0x71, (byte)0x09, (byte)0x0A,
        (byte)0x97, (byte)0x4C, (byte)0xE7, (byte)0x13, (byte)0xE7, (byte)0x69,
        (byte)0x28, (byte)0x91, (byte)0xD5, (byte)0xC5, (byte)0x55, (byte)0x5A,
        (byte)0x55, (byte)0xB2, (byte)0xAC, (byte)0xAD, (byte)0x97, (byte)0x8C,
        (byte)0x26, (byte)0x9B, (byte)0x0D, (byte)0x05, (byte)0x9C, (byte)0xD0,
        (byte)0xB0, (byte)0xEC, (byte)0xF0, (byte)0xA6, (byte)0xCE, (byte)0x5F,
        (byte)0xDD, (byte)0x2D, (byte)0xC6, (byte)0x9D, (byte)0x11, (byte)0x6E,
        (byte)0xA0, (byte)0x3A, (byte)0x04, (byte)0x10, (byte)0x0A, (byte)0x15,
        (byte)0xB3, (byte)0x05, (byte)0x45, (byte)0x82, (byte)0x7A, (byte)0x80,
        (byte)0x14, (byte)0x09, (byte)0x00, (byte)0x21, (byte)0xFE, (byte)0x4F,
        (byte)0x43, (byte)0x6F, (byte)0x70, (byte)0x79, (byte)0x72, (byte)0x69,
        (byte)0x67, (byte)0x68, (byte)0x74, (byte)0x20, (byte)0x32, (byte)0x30,
        (byte)0x30, (byte)0x30, (byte)0x20, (byte)0x62, (byte)0x79, (byte)0x20,
        (byte)0x53, (byte)0x75, (byte)0x6E, (byte)0x20, (byte)0x4D, (byte)0x69,
        (byte)0x63, (byte)0x72, (byte)0x6F, (byte)0x73, (byte)0x79, (byte)0x73,
        (byte)0x74, (byte)0x65, (byte)0x6D, (byte)0x73, (byte)0x2C, (byte)0x20,
        (byte)0x49, (byte)0x6E, (byte)0x63, (byte)0x2E, (byte)0x20, (byte)0x41,
        (byte)0x6C, (byte)0x6C, (byte)0x20, (byte)0x52, (byte)0x69, (byte)0x67,
        (byte)0x68, (byte)0x74, (byte)0x73, (byte)0x20, (byte)0x52, (byte)0x65,
        (byte)0x73, (byte)0x65, (byte)0x72, (byte)0x76, (byte)0x65, (byte)0x64,
        (byte)0x2E, (byte)0x0D, (byte)0x0A, (byte)0x4A, (byte)0x4C, (byte)0x46,
        (byte)0x20, (byte)0x47, (byte)0x52, (byte)0x20, (byte)0x56, (byte)0x65,
        (byte)0x72, (byte)0x20, (byte)0x31, (byte)0x2E, (byte)0x30, (byte)0x0D,
        (byte)0x0A, (byte)0x00, (byte)0x3B
    };

    /**
     * The redo icon.
     */
    public static final Icon REDO_ICON = new ImageIcon(
        TOOLKIT.createImage(REDO_ICON_BYTES),
        "Redo"
        );

    private AbstractAction2 undoAction, redoAction;
    
    private boolean needsUpdate;

    /**
     * Creates a new ActionedUndoManager.
     */
    public ActionedUndoManager()
    {
        super();

        undoAction = new UndoAction();
        redoAction = new RedoAction();
        needsUpdate = false;
        undoAction.setEnabled(canUndo());
        redoAction.setEnabled(canRedo());
    }

    /**
     * Returns an action that calls the undo method of this ActionedUndoManager.  This
     * ActionedUndoManager also manages the enabled status of the action.
     *
     * @return The action for the undo operation.
     */
    public AbstractAction2 getUndoAction()
    {
        return undoAction;
    }

    /**
     * Returns an action that calls the redo method of this ActionedUndoManager.  This
     * ActionedUndoManager also manages the enabled status of the action.
     *
     * @return The action for the redo operation.
     */
    public AbstractAction2 getRedoAction()
    {
        return redoAction;
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public void discardAllEdits()
    {
        needsUpdate = true;

        super.discardAllEdits();

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    protected void trimForLimit()
    {
        needsUpdate = true;

        super.trimForLimit();

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    protected void trimEdits(int from, int to)
    {
        needsUpdate = true;

        super.trimEdits(from, to);

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public void setLimit(int l)
    {
        needsUpdate = true;

        super.setLimit(l);

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }


    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    protected void undoTo(UndoableEdit edit) throws CannotUndoException
    {
        needsUpdate = true;

        super.undoTo(edit);

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    protected void redoTo(UndoableEdit edit) throws CannotRedoException
    {
        needsUpdate = true;

        super.redoTo(edit);

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public void undoOrRedo() throws CannotRedoException, CannotUndoException
    {
        needsUpdate = true;

        super.undoOrRedo();

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public void undo() throws CannotUndoException
    {
        needsUpdate = true;

        super.undo();

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public void redo() throws CannotRedoException
    {
        needsUpdate = true;

        super.redo();

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public boolean addEdit(UndoableEdit anEdit)
    {
        needsUpdate = true;

        boolean returnValue = super.addEdit(anEdit);

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }

        return returnValue;
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public void end()
    {
        needsUpdate = true;

        super.end();

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public void undoableEditHappened(UndoableEditEvent e)
    {
        needsUpdate = true;

        super.undoableEditHappened(e);

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }
    }

    /**
     * Intercept all calls that could affect the undo/redo action states.
     */
    public boolean replaceEdit(UndoableEdit anEdit)
    {
        needsUpdate = true;

        boolean returnValue = super.replaceEdit(anEdit);

        if (needsUpdate)
        {
            needsUpdate = false;
            undoAction.setEnabled(canUndo());
            redoAction.setEnabled(canRedo());
        }

        return returnValue;
    }

    /**
     * UndoAction is an action which calls the undo() method of its ActionedUndoManager.  This
     * action's ActionedUndoManager also manages its enabled status.
     *
     * @author btsang
     * @version 7.1
     */
    protected class UndoAction extends AbstractAction2
    {
        /**
         * Creates a new UndoAction.
         */
        protected UndoAction()
        {
            super("Undo", UNDO_ICON);

            setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z, TOOLKIT.getMenuShortcutKeyMask()));

            setMnemonic(KeyEvent.VK_U);
        }

        /**
         * Implements ActionListener to perform the undo operation.
         */
        public void actionPerformed(ActionEvent e)
        {
            undo();
        }
    }

    /**
     * RedoAction is an action which calls the redo() method of its ActionedUndoManager.  This
     * action's ActionedUndoManager also manages its enabled status.
     *
     * @author btsang
     * @version 7.1
     */
    protected class RedoAction extends AbstractAction2
    {
        /**
         * Creates a new RedoAction.
         */
        protected RedoAction()
        {
            super("Redo", REDO_ICON);

            setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Y, TOOLKIT.getMenuShortcutKeyMask()));

            setMnemonic(KeyEvent.VK_R);
        }

        /**
         * Implements ActionListener to perform the redo operation.
         */
        public void actionPerformed(ActionEvent e)
        {
            redo();
        }
    }
}