package edu.princeton.swing;

import java.io.File;
import javax.swing.filechooser.*;
import java.util.Iterator;
import java.util.Vector;

/**
 * ExtensionFileFilter extends FileFilter to provide a generalized extension-based file filter.
 */
public class ExtensionFileFilter extends FileFilter
{
    private String[] extensions;
    private int extensionCount;
    private String description;
    private String baseDescription;

    /**
     * Creates a new ExtensionFileFilter.
     */
    public ExtensionFileFilter()
    {
        this(null, "");
    }

    /**
     * Creates a new ExtensionFileFilter.
     */
    public ExtensionFileFilter(String baseDescription)
    {
        this(null, baseDescription);
    }

    /**
     * Creates a new ExtensionFileFilter.
     */
    public ExtensionFileFilter(String extension, String baseDescription)
    {
        if (baseDescription == null)
            throw new NullPointerException();
        this.baseDescription = baseDescription;

        extensions = new String[10];
        if (extension != null)
        {
            extensions[0] = "." + extension;
            extensionCount = 1;
        }
        else {
            extensionCount = 0;
        }
    }

    /**
     * Extend FileFilter to accept only files with an extension matching one added to the filter.
     */
    public boolean accept(File file)
    {
        if (file.isDirectory())
            return true;

        String fileName = file.getName();

        for (int ctr = 0; ctr < extensionCount; ctr++)
        {
            if (fileName.endsWith(extensions[ctr]))
                return true;
        }

        return false;
    }

    /**
     * Return the number of extensions this filter will accept.
     *
     * @return The number of extensions this filter will accept.
     */
    public int getExtensionCount()
    {
        return extensionCount;
    }

    /**
     * Return the extension at a given index.
     *
     * @param index The index of the desired extension.  An ArrayIndexOutOfBoundsException will
     * be thrown if the index is invalid.
     * @return The extension at a given index.
     */
    public String getExtension(int index)
    {
        if (index < 0 || index >= extensionCount)
            throw new ArrayIndexOutOfBoundsException();

        return extensions[index].substring(1);
    }

    /**
     * Adds an extension to the list of accepted extensions.
     *
     * @param extension The new extension to accept.  A null value will result in a
     * NullPointerException.
     */
    public void addExtension(String extension)
    {
        if (extension == null)
            throw new NullPointerException();

        if (extensionCount == extensions.length)
        {
            String oldExtensions[] = extensions;
            extensions = new String[extensionCount * 2 + 1];

            for (int ctr = 0; ctr < extensionCount; ctr++)
                extensions[ctr] = oldExtensions[ctr];
        }

        extensions[extensionCount] = "." + extension;
        extensionCount++;

        description = null;
    }

    /**
     * Removes all extensions from the list of accepted extensions.
     */
    public void clearExtensions()
    {
        extensionCount = 0;
        description = null;
    }

    /**
     * Gets a description of this filter.
     *
     * @return A description of this filter, eg.
     * &quot;Web Image Files (*.jpg;*.jpeg;*.gif;*.png)&quot;
     */
    public String getDescription()
    {
        if (description == null)
        {
            StringBuffer buffer = new StringBuffer(baseDescription);
            buffer.append(" (");
            for (int ctr = 0; ctr < extensionCount; ctr++)
            {
                buffer.append('*');
                buffer.append(extensions[ctr]);
                if (ctr < extensionCount - 1)
                    buffer.append(';');
            }
            buffer.append(')');

            description = buffer.toString();
        }

        return description;
    }

    /**
     * Sets the base description of this filter.
     *
     * @param baseDescription The base description of this filter, eg. &quot;Web Image Files&quot;
     */
    public void setBaseDescription(String baseDescription)
    {
        if (baseDescription == null)
            throw new NullPointerException();
        this.baseDescription = baseDescription;
    }
}
