package edu.princeton.swing;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;

import javax.swing.*;
import javax.swing.text.*;

/**
 * PTextArea extends JTextArea to provide ClipboardTarget support.
 *
 * @author btsang
 * @version 7.1
 */

public class PTextArea extends JTextArea implements ClipboardTarget
{
    private static final Toolkit TOOLKIT = Toolkit.getDefaultToolkit();

    private JPopupMenu popupMenu;

    /**
     * Constructs a new PTextArea.
     */
    public PTextArea()
    {
        super();

        initTextArea();
    }

    /**
     * Constructs a new PTextArea.
     */
    public PTextArea(String text)
    {
        super(text);

        initTextArea();
    }

    /**
     * Constructs a new PTextArea.
     */
    public PTextArea(int rows, int columns)
    {
        super(rows, columns);

        initTextArea();
    }

    /**
     * Constructs a new PTextArea.
     */
    public PTextArea(String text, int rows, int columns)
    {
        super(text, rows, columns);

        initTextArea();
    }

    /**
     * Constructs a new PTextArea.
     */
    public PTextArea(Document doc)
    {
        super(doc);

        initTextArea();
    }

    /**
     * Constructs a new PTextArea.
     */
    public PTextArea(Document doc, String text, int rows, int columns)
    {
        super(doc, text, rows, columns);

        initTextArea();
    }

    /**
     * Called by the constructors to init the TList properly.
     */
    private void initTextArea()
    {
        popupMenu = ClipboardTargetManager.createDefaultPopupMenu();

        ClipboardTargetManager.installMappings(this);

        enableEvents(AWTEvent.FOCUS_EVENT_MASK | AWTEvent.MOUSE_EVENT_MASK);
    }

    /**
     * Returns wheter or not the cut operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to cut() should be permitted.
     */
    public boolean canCut()
    {
        return isEnabled() && isEditable();
    }

    /**
     * Returns wheter or not the copy operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to copy() should be permitted.
     */
    public boolean canCopy()
    {
        return isEnabled();
    }

    /**
     * Returns wheter or not the paste operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to paste() should be permitted.
     */
    public boolean canPaste()
    {
        return isEnabled() && isEditable();
    }

    /**
     * Returns wheter or not the selectAll operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to selectAll() should be permitted.
     */
    public boolean canSelectAll()
    {
        return isEnabled();
    }

    /**
     * Intercept FocusEvents to notify the ClipboardTargetManager of the event.
     */
    protected void processFocusEvent(FocusEvent e)
    {
        if (e.getID() == FocusEvent.FOCUS_GAINED)
        {
            ClipboardTargetManager.targetGainedFocus(this);
        }
        else if (e.getID() == FocusEvent.FOCUS_LOST)
        {
            ClipboardTargetManager.targetLostFocus(this);
        }

        super.processFocusEvent(e);
    }

    /**
     * Intercept MouseEvents to activate the popupMenu at the right time.
     */
    protected void processMouseEvent(MouseEvent e)
    {
        if (!isEnabled())
            return;

        boolean passedToSuper = false;

        if (popupMenu != null && e.isPopupTrigger())
        {
            if (!hasFocus())
                requestFocus();

            int x = e.getX();
            int y = e.getY();
            Point componentLocation = getLocationOnScreen();
            Dimension screenSize = TOOLKIT.getScreenSize();
            Dimension preferredSize = popupMenu.getPreferredSize();

            if (componentLocation.x + x + preferredSize.width > screenSize.width)
                x -= preferredSize.width;

            if (componentLocation.y + y + preferredSize.height > screenSize.height)
                y -= preferredSize.height;

            e.consume();
            passedToSuper = true;

            super.processMouseEvent(e);

            popupMenu.show(this, x, y);
        }

        if (!passedToSuper)
            super.processMouseEvent(e);
    }

    /**
     * Override the setEnabled() method to intercept any changes concerning the enabled status of
     * the component.
     */
    public void setEnabled(boolean enabled)
    {
        super.setEnabled(enabled);

        ClipboardTargetManager.targetAbilityChanged(this);
    }

    /**
     * Override the setEditable() method to intercept any changes concerning the editable status of
     * the component.
     */
    public void setEditable(boolean editable)
    {
        super.setEditable(editable);

        ClipboardTargetManager.targetAbilityChanged(this);
    }

    /**
     * Gets the popup menu for this component.
     *
     * @return The popup menu for the component.  Null is returned if this component does not have
     * a popup menu.
     */
    public JPopupMenu getPopupMenu()
    {
        return popupMenu;
    }

    /**
     * Sets the popup menu for this component.
     *
     * @param popupMenu The popup menu for the component.  A null value indicates that this
     * component should not have a popup menu.
     */
    public void setPopupMenu(JPopupMenu popupMenu)
    {
        if (popupMenu != this.popupMenu)
        {
            if (popupMenu != null && this.popupMenu.isVisible())
            {
                this.popupMenu.setVisible(false);
            }

            JPopupMenu oldValue = this.popupMenu;
            this.popupMenu = popupMenu;

            firePropertyChange("popupMenu", oldValue, popupMenu);
        }
    }
}
